/// @addtogroup bitSet
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file EuclBitSet_BitParIterator.h
///
/// @author Marian Mrozek
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2006 by the CAPD Group.
//
// This file constitutes a part of the CAPD library,
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details.

/*=======================================================================*/
/*  ****************** Not tested!!! Not working!!! ******************** */
/*=======================================================================*/

// BitParIterator is a general use iterator for bitmap sets
// It behaves similarly to the BitIterator but
// unlike the previous operators it stores and updates the
// parity of cooridnates of the pixel

//  ************** under construction ****************

#if !defined(_EUCLBITSETT_BITPARITERATOR_H_)
#define _EUCLBITSETT_BITPARITERATOR_H_

template <typename P_BitSet,int dim>
class EuclBitSetT<P_BitSet,dim>::BitParIterator : public EuclBitSetT<P_BitSet,dim>::BitIterator{
  typedef typename EuclBitSetT<P_BitSet,dim>::BitIterator BitIterator;
  friend class EuclBitSetT<P_BitSet,dim>;
protected:
  static const int bitsPerWord = 8*sizeof(typename EuclBitSetT<P_BitSet,dim>::Word);
  unsigned int parity;


  void setParityFromPosition(int pos){
    int c;
    parity=0;
    for(int i=0;i<dim;++i){
      c=pos % this->baseEuclBitSet()->getPaddedWidth(i);
      pos /= this->baseEuclBitSet()->getPaddedWidth(i);
      parity |= ( (c & 1U) << i);
    }
  }

  bool findPoint(){
    this->BitIterator::findPoint();
    int pos=this->position();
    setParityFromPosition(pos);
    return this->wIt < this->baseEuclBitSet()->bitmapEnd;
  }



public:

  BitParIterator(const EuclBitSetT& s):BitIterator(s){
    parity=0;
  }

  // Constructor of an image of iterator A_it in the bitmap s
  BitParIterator(const EuclBitSetT& s,const BitParIterator& A_it):
    BitIterator(s,A_it),
    parity(A_it.parity)
  {}


  BitParIterator(const EuclBitSetT& s,const int* A_coord):BitIterator(s,A_coord){
    parity=0;
    for(int k=0;k<dim;++k){
      parity |= ( (A_coord[k] & 1U) << k);
    }
  }

  BitParIterator(const EuclBitSetT& s, Word pos):BitIterator(s,pos){
    setParityFromPosition(pos);
  }

  // intended as a converter
  BitParIterator(const BitIterator& A_it): BitIterator(A_it){
    int pos=this->position();
    setParityFromPosition(pos);
  }

  template<typename P_CubSetIndex>
  BitParIterator(const EuclBitSetT& A_set,const P_CubSetIndex& A_index): BitIterator(A_set,A_index){
    int pos=this->position();
    setParityFromPosition(pos);
  }

  bool odd(int i){
    return (parity >> i) & 1U;
  }

  BitParIterator& decInDir(int i){
    parity ^= (1 << i);
    if(i==0){
      BitIterator::operator--();
    }else{
      this->wIt -= this->baseEuclBitSet()->wordStride[i];
    }
    return *this;
  }

  BitParIterator& incInDir(int i){
    parity ^= (1 << i);
    if(i==0){
      BitIterator::operator++();
    }else{
      this->wIt += this->baseEuclBitSet()->wordStride[i];
    }
    return *this;
  }

  BitParIterator& operator++(){
    BitIterator::operator++();
    int pos=this->position();
    setParityFromPosition(pos);
    return *this;
  }

  BitParIterator& operator--(){
    BitIterator::operator--();
    int pos=this->position();
    setParityFromPosition(pos);
    return *this;
  }

  int operator[](int j) const{
    int pos=this->position();
    int c=0;
    for(int i=0;i<=j;++i){
      c=pos % this->baseEuclBitSet()->getPaddedWidth(i);
      pos /= this->baseEuclBitSet()->getPaddedWidth(i);
    }
    return c;
  }

/* This will not work
  const int* coords() const{
    return coord;
  }
*/

  int ownDim() const{
    static int* table=0;
    static unsigned int length=0;

    while(parity>=length){
      if(length>0){
        int newlength=2*length;
        int* newtable=new int[newlength];
        for(unsigned int i=0;i<length;++i){
          newtable[i]=table[i];
          newtable[length+i]=1+table[i];
        }
        delete[] table;
        table=newtable;
        length=newlength;
      }else{
        length=2;
        table=new int[length];
        table[0]=0;
        table[1]=1;
      }
    }

    return table[parity];
  }

  int embDim() const{
    return dim;
  }

  Pixel operator*(){
    int coords[dim];
    this->getCoords(coords);
    return Pixel(coords);
  }

  std::string toString(){
    std::ostringstream s;
    s << "BitParIter for bitmap at " << this->baseEuclBitSet() << " with wIt=" << this->wIt << " bit=" << this->bit << " at pixel " << *this << std::endl;
    return s.str();
  }

  friend std::ostream& operator<<(std::ostream& out,const BitParIterator& A_it){
    int coords[dim];
    A_it.getCoords(coords);
    out << "BitCoordIt at(" << coords[0];
    for(int i=1;i<dim;++i) out << "," << coords[i];
    out << ")[" << A_it.position() << "]" ;
    return out;
  }

};
#endif //_EUCLBITSETT_BITPARITERATOR_H_

/*=======================================================================*/
/// @}

