//////////////////////////////////////////////////////////////////////////////
//   Package:          CAPD

/////////////////////////////////////////////////////////////////////////////
//
/// @file TypeTraits.h
///
/// @author Tomasz Kapela   @date 2010-03-08
//
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) Tomasz Kapela 2010
//
// This file constitutes a part of the CAPD library,
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.ii.uj.edu.pl/ for details.

#ifndef _CAPD_CAPD_TYPETRAITS_H_
#define _CAPD_CAPD_TYPETRAITS_H_

#include "capd/interval/Interval.h"
#include "capd/vectalg/Z2.h"
#include <limits>

namespace capd {

/**
 *  Defines type traits such as their values zero, one etc.
 */
template <typename T>
class TypeTraits {
public:
  /// returns object set to zero
  static inline T zero(){
    return static_cast<T>(0.0);
  }
  /// returns object set to one
  static inline T one(){
    return static_cast<T>(1.0);
  }
  /// number of decimal digits
  static inline int numberOfDigits(){
    return std::numeric_limits<T>::digits10;
  }
};

/**
 * Specialization for intervals
 */
template <typename T, typename R>
class TypeTraits < capd::intervals::Interval<T,R> > {
public:
  static inline const capd::intervals::Interval<T,R> & zero(){
    return S_zero;
  }
  static inline const capd::intervals::Interval<T,R> & one(){
    return S_one;
  }
  static inline int numberOfDigits(){
     return TypeTraits<T>::numberOfDigits();
   }
private:
  static const  capd::intervals::Interval<T,R> S_zero ;// = capd::intervals::Interval<T,R>(T(0.0));
  static const  capd::intervals::Interval<T,R> S_one ;
};

template <typename T, typename R>
const capd::intervals::Interval<T,R> TypeTraits<capd::intervals::Interval<T,R> >::S_zero = capd::intervals::Interval<T,R>(
    TypeTraits<T>::zero()
);

template <typename T, typename R>
const capd::intervals::Interval<T,R> TypeTraits<capd::intervals::Interval<T,R> >::S_one = capd::intervals::Interval<T,R>(
    TypeTraits<T>::one()
);

/**
 * Traits of type int
 */
template <>
class TypeTraits<int> {
public:
  static inline int zero(){
    return 0;
  }
  static inline int one(){
    return 1;
  }
  static inline int numberOfDigits(){
      return std::numeric_limits<int>::digits10;
  }
};

/**
 * Traits of type Z2
 */
template <>
class TypeTraits<Z2> {
public:
  static inline int zero(){
    return 0;
  }
  static inline int one(){
    return 1;
  }
  static inline int numberOfDigits(){
      return 1;
  }
};

} // end of namespace capd

#endif // _CAPD_CAPD_TYPETRAITS_H_
